@@ -3,8 +3,7 @@ use core::marker;
 
 use crate::integer_encoding::{IntegerEncoding, UsizeEncoding};
 use musli::de::{
-    Decoder, MapDecoder, MapEntryDecoder, PackDecoder, PairDecoder, ReferenceVisitor,
-    SequenceDecoder, StructDecoder,
+    Decoder, PackDecoder, PairDecoder, PairsDecoder, ReferenceVisitor, SequenceDecoder,
 };
 use musli::error::Error;
 use musli_binary_common::reader::PositionedReader;
@@ -63,13 +62,18 @@ where
     type Tuple = RemainingSimpleDecoder<R, I, L>;
     type Variant = Self;
 
+    #[inline]
+    fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "type supported by the storage decoder")
+    }
+
     #[inline]
     fn decode_unit(mut self) -> Result<(), Self::Error> {
         let pos = self.reader.pos();
         let count = L::decode_usize(&mut self.reader)?;
 
         if count != 0 {
-            return Err(Self::Error::collect_from_display(ExpectedEmptySequence {
+            return Err(Self::Error::message(ExpectedEmptySequence {
                 actual: count,
                 pos,
             }));
@@ -94,8 +98,7 @@ where
         V: ReferenceVisitor<'de, Target = [u8], Error = Self::Error>,
     {
         let len = L::decode_usize(&mut self.reader)?;
-        let bytes = self.reader.read_bytes(len)?;
-        visitor.visit_ref(bytes)
+        self.reader.read_bytes(len, visitor)
     }
 
     #[inline]
@@ -116,20 +119,20 @@ where
             type Error = V::Error;
 
             #[inline]
-            fn expected(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-                self.0.expected(f)
+            fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+                self.0.expecting(f)
             }
 
             #[inline]
-            fn visit_ref(self, bytes: &'de [u8]) -> Result<Self::Ok, Self::Error> {
+            fn visit_borrowed(self, bytes: &'de [u8]) -> Result<Self::Ok, Self::Error> {
                 let string = core::str::from_utf8(bytes).map_err(Self::Error::custom)?;
-                self.0.visit_ref(string)
+                self.0.visit_borrowed(string)
             }
 
             #[inline]
-            fn visit(self, bytes: &[u8]) -> Result<Self::Ok, Self::Error> {
+            fn visit_any(self, bytes: &[u8]) -> Result<Self::Ok, Self::Error> {
                 let string = core::str::from_utf8(bytes).map_err(Self::Error::custom)?;
-                self.0.visit(string)
+                self.0.visit_any(string)
             }
         }
     }
@@ -142,10 +145,7 @@ where
         match byte {
             0 => Ok(false),
             1 => Ok(true),
-            b => Err(Self::Error::collect_from_display(BadBoolean {
-                actual: b,
-                pos,
-            })),
+            b => Err(Self::Error::message(BadBoolean { actual: b, pos })),
         }
     }
 
@@ -156,10 +156,7 @@ where
 
         match char::from_u32(num) {
             Some(d) => Ok(d),
-            None => Err(Self::Error::collect_from_display(BadCharacter {
-                actual: num,
-                pos,
-            })),
+            None => Err(Self::Error::message(BadCharacter { actual: num, pos })),
         }
     }
 
@@ -291,7 +288,7 @@ where
     }
 
     #[inline]
-    fn finish(self) -> Result<(), Self::Error> {
+    fn end(self) -> Result<(), Self::Error> {
         Ok(())
     }
 }
@@ -316,7 +313,7 @@ where
     L: UsizeEncoding,
 {
     type Error = R::Error;
-    type Next<'this> = StorageDecoder<&'this mut R, I, L> where Self: 'this;
+    type Decoder<'this> = StorageDecoder<&'this mut R, I, L> where Self: 'this;
 
     #[inline]
     fn size_hint(&self) -> Option<usize> {
@@ -324,7 +321,7 @@ where
     }
 
     #[inline]
-    fn decode_next(&mut self) -> Result<Option<Self::Next<'_>>, Self::Error> {
+    fn next(&mut self) -> Result<Option<Self::Decoder<'_>>, Self::Error> {
         if self.remaining == 0 {
             return Ok(None);
         }
@@ -334,7 +331,7 @@ where
     }
 }
 
-impl<'de, R, I, L> MapDecoder<'de> for RemainingSimpleDecoder<R, I, L>
+impl<'de, R, I, L> PairsDecoder<'de> for RemainingSimpleDecoder<R, I, L>
 where
     R: PositionedReader<'de>,
     I: IntegerEncoding,
@@ -342,7 +339,7 @@ where
 {
     type Error = R::Error;
 
-    type Entry<'this> = StorageDecoder<&'this mut R, I, L>
+    type Decoder<'this> = StorageDecoder<&'this mut R, I, L>
     where
         Self: 'this;
 
@@ -352,56 +349,7 @@ where
     }
 
     #[inline]
-    fn decode_entry(&mut self) -> Result<Option<Self::Entry<'_>>, Self::Error> {
-        if self.remaining == 0 {
-            return Ok(None);
-        }
-
-        self.remaining -= 1;
-        Ok(Some(StorageDecoder::new(&mut self.decoder.reader)))
-    }
-}
-
-impl<'de, R, I, L> MapEntryDecoder<'de> for StorageDecoder<R, I, L>
-where
-    R: PositionedReader<'de>,
-    I: IntegerEncoding,
-    L: UsizeEncoding,
-{
-    type Error = R::Error;
-    type Key<'this> = StorageDecoder<&'this mut R, I, L> where Self: 'this;
-    type Value<'this> = StorageDecoder<&'this mut R, I, L> where Self: 'this;
-
-    #[inline]
-    fn decode_key(&mut self) -> Result<Self::Key<'_>, Self::Error> {
-        Ok(StorageDecoder::new(&mut self.reader))
-    }
-
-    #[inline]
-    fn decode_value(&mut self) -> Result<Self::Value<'_>, Self::Error> {
-        Ok(StorageDecoder::new(&mut self.reader))
-    }
-}
-
-impl<'de, R, I, L> StructDecoder<'de> for RemainingSimpleDecoder<R, I, L>
-where
-    R: PositionedReader<'de>,
-    I: IntegerEncoding,
-    L: UsizeEncoding,
-{
-    type Error = R::Error;
-
-    type Field<'this> = StorageDecoder<&'this mut R, I, L>
-    where
-        Self: 'this;
-
-    #[inline]
-    fn size_hint(&self) -> Option<usize> {
-        Some(self.remaining)
-    }
-
-    #[inline]
-    fn decode_field(&mut self) -> Result<Option<Self::Field<'_>>, Self::Error> {
+    fn next(&mut self) -> Result<Option<Self::Decoder<'_>>, Self::Error> {
         if self.remaining == 0 {
             return Ok(None);
         }
@@ -422,12 +370,12 @@ where
     type Second = StorageDecoder<R, I, L>;
 
     #[inline]
-    fn decode_first(&mut self) -> Result<Self::First<'_>, Self::Error> {
+    fn first(&mut self) -> Result<Self::First<'_>, Self::Error> {
         Ok(StorageDecoder::new(&mut self.reader))
     }
 
     #[inline]
-    fn decode_second(self) -> Result<Self::Second, Self::Error> {
+    fn second(self) -> Result<Self::Second, Self::Error> {
         Ok(StorageDecoder::new(self.reader))
     }
 
